#pragma once
#include <stdint.h>
#include <stdio.h>

#include "KissSecurityManager.h"

typedef enum PROCESS_PHASE {
	PROCESS_DISABLED			= 0,
	SYNC_PHASE					= 1,
	IDLE_PHASE					= 2,
	SEND_PHASE					= 3,
	LISTEN_PHASE				= 4,
	PROCESS_PHASE_BITS			= 3,
	PROCESS_PHASE_MASK			= ((1 << PROCESS_PHASE_BITS) - 1)
} PROCESS_PHASE;

enum {
	USER_STREAM_BYTE			= 1,
	CTRL_TIMESYNC				= CTRL_FRAME_TYPE_BASE + 1,
	CTRL_REGISTER				= CTRL_FRAME_TYPE_BASE + 2,
	CTRL_UNREGISTER				= CTRL_FRAME_TYPE_BASE + 3,
	CTRL_OPEN_CONNECTION		= CTRL_FRAME_TYPE_BASE + 4,
	CTRL_ACK					= CTRL_FRAME_TYPE_BASE + 5,
	CTRL_ADVERTISEMENT			= CTRL_FRAME_TYPE_BASE + 6,
	CTRL_LANTERN				= CTRL_FRAME_TYPE_BASE + 7,
	CTRL_ERROR					= CTRL_FRAME_TYPE_BASE + 8,
	CTRL_HEAD_LIST				= CTRL_FRAME_TYPE_BASE + 9,
	CTRL_FRAME_TYPE_MASK		= ((1 << FRAME_TYPE_BITS) - 1)
} CTRL_FRAME_TYPE;

enum {
	GENERATED_FRAME		= 1,
	OMITTED_FRAME		= 2,
	WAIT4PROCESS		= 4,
	END_OF_SESSION		= 8,
	EMULATED			= 16,
	FRAME_SOURCE_MASK	= 0x1F
} FRAME_SOURCE;

typedef struct
{
	uint8_t len;
	uint8_t bytes[8];
} __attribute__((packed)) MAC_ADDR;

// first byte of header not encrypted
typedef struct
{
	uint8_t			srcRouter;	// field never signed/encrypted
	NODE_ID			srcNode;
	NODE_ID			dstNode;
	uint8_t			frameType:FRAME_TYPE_BITS;
	uint8_t			response:1;		// this is frame response 
	uint8_t			noResponse:1;	// this is command with no response
	uint8_t			emulated:1;
	uint8_t			userMultiResponse:1;
	uint8_t			pan;			// network identifier (Personal Area Network)
	uint8_t			dummy:4;
	uint8_t			netStatus:4;
	uint8_t			frameSeq;		// current command/response ID
	uint8_t			frameLen;		// current frame length
} __attribute__((packed)) FHEAD;

static const NODE_ID	nullNodeId;

typedef struct
{
	uint8_t		bodyLen;			// real length of body
	uint8_t		bodyType;
#define	PAYLOAD_HEADER_SIZE	2
	uint8_t		body[MAX_PACKET_SIZE - sizeof(FHEAD) - PAYLOAD_HEADER_SIZE - 
					KissSecurityManager::OPEN_HEAD_SIZE - KissSecurityManager::HIDDEN_HEAD_SIZE];
} __attribute__((packed)) PAYLOAD;
#undef PAYLOAD_HEADER_SIZE

static const uint8_t MAX_BODY_SIZE = sizeof(((PAYLOAD *) 0)->body);
static const uint8_t PAYLOAD_HEADER_SIZE = sizeof(PAYLOAD) - MAX_BODY_SIZE;

typedef struct
{
	uint8_t		openHead[KissSecurityManager::OPEN_HEAD_SIZE];
	FHEAD		shead;		// sender header
	uint8_t		hiddenHead[KissSecurityManager::HIDDEN_HEAD_SIZE];
	PAYLOAD		payload;	// frame body
} __attribute__((packed)) FRAME;

#undef OPEN_SECURITY_HEAD
#undef HIDDEN_SECURITY_HEAD

static const uint8_t MIN_FRAME_LENGTH = offsetof(FRAME, payload) + PAYLOAD_HEADER_SIZE;

typedef struct 
{
	uint64_t	startTime;		// send/receive start time
	uint64_t	endTime;		// send/receive end time
	uint8_t		len;			// length of  oryginal packet (encrypted/decrypted)
	uint8_t		rssi;
	int8_t		channel;
	int8_t		cmdIdx;
	bool		timeSchema;		// F - local/T-router
} __attribute__((packed)) IFRAME;

typedef struct
{
	FHEAD		head;			// cmd/response buffer
	uint16_t	startInt;		// registration/receive interval
	uint16_t	frameSendInt;	// frame or ACK sended
	uint16_t	ackSendInt;		// ACK send interval. For delayed remove
	uint8_t		codedLen;		// length of coded packet
	uint8_t		poolHandle;		// index of frame in memory pool
	uint8_t		wait4response:1;// this frame wait for response frame (not only for ACK)
	uint8_t		ignored:1;
	uint8_t		ackReceived:1;	// frame accepted by node. Not response yet!
	uint8_t		endOfSession:1;
	uint8_t		changed:1;
	uint8_t		dummy:3;
	
	uint8_t		source:1;		// source 0-Received from net, 1-generated by node
	uint8_t		omitted:1;
	uint8_t		wait4process:1;
	uint8_t		wait4handler:1;	// frame out of sequences. Wait for absent packet
	uint8_t		virtualOwner:NODE_ID_BITS;	// virtual node no or ZERO
} __attribute__((packed)) COMMAND_STATE;

typedef struct 
{
	int8_t			slotChannel:CHANNELS_IDX_BITS;
	PROCESS_PHASE	phase:PROCESS_PHASE_BITS;
} __attribute__((packed)) TIME_SLOT_MODE;

//----------------------------- CTRL FRAMES --------------------------------
// Router advertisement
typedef struct
{
	COMM_SCHEMA		routerSchema;
	uint8_t			canRegister:1;		// router can register regular nodes
	uint8_t			registerActive:1;	// register gap active in extra time sub-interval
	uint8_t			reservePriority:2;
	uint8_t			routerGap:3;
	uint8_t			dummy:1;
	uint16_t		nodes2wait;			// bitmap of nodes (devices) which should 
										// wait for frame in current transmit interval
	uint16_t		nodes2send;			// bitmap of nodes (devices!) enabled to transmit 
										// in nearest router listen sub-interval
	uint16_t		extraTimeNodes;		// how to use extra time in transmit sub-interval
	uint64_t		routerTime;			// for time sync (temporary)
	uint32_t		eosId;
} __attribute__((packed)) CTRL_RADVERT_BODY;
static const short ADVERTISEMENT_SIZE =	MIN_FRAME_LENGTH + sizeof(CTRL_RADVERT_BODY);
static const short POOL_CHUNK_SIZE = ((ADVERTISEMENT_SIZE - 1) / 8) * 8;

// Device advertisement
typedef struct
{
	COMM_SCHEMA		routerSchema;
	uint64_t		routerTime;			// for time sync (temporary)
	uint8_t			reservePriority:2;
	uint8_t			dummy:6;
} __attribute__((packed)) CTRL_LANTERN_BODY;

typedef struct
{
	FHEAD			errroOrigin;
	int16_t			errorCode;
	uint8_t			errorDescType;
	uint8_t			errorDescLen;
	uint8_t			errorDesc[];
} __attribute__((packed)) CTRL_ERROR_BODY;

typedef struct
{
	uint8_t			headListLen;
	FHEAD			headList[];
} __attribute__((packed)) HEAD_LIST;

static const uint8_t ACK_LIST_SIZE = (MAX_BODY_SIZE - sizeof(HEAD_LIST)) / sizeof(FHEAD);

typedef struct
{
	uint64_t			time;
} __attribute__((packed)) CTRL_TIMESYNC_FRAME;

typedef struct
{
	MAC_ADDR		mac;
	MAC_ADDR		routerMAC;
} __attribute__((packed)) CTRL_REGISTER_BODY;

typedef struct
{
	MAC_ADDR		srcMac;
	MAC_ADDR		dstMac;
	
	uint8_t			dummy:6;
	uint8_t			port:CONNECTION_ID_BITS;
} __attribute__((packed)) CTRL_CONNECTION_BODY;
